from argparse import ArgumentParser, ArgumentTypeError
from locale import getdefaultlocale
from multiprocessing import Pool
from contextlib import redirect_stdout
from io import StringIO

from zdict import constants, utils, easter_eggs
from zdict.api import dump
from zdict.completer import DictCompleter
from zdict.loader import get_dictionary_map
from zdict.utils import readline, check_zdict_dir_and_db


def user_set_encoding_and_is_utf8():
    # Check user's encoding settings
    try:
        (lang, enc) = getdefaultlocale()
    except ValueError:
        print("Didn't detect your LC_ALL environment variable.")
        print("Please export LC_ALL with some UTF-8 encoding.")
        return False
    else:
        if enc != "UTF-8":
            print("zdict only works with encoding=UTF-8, ")
            print("but your encoding is: {} {}".format(lang, enc))
            print("Please export LC_ALL with some UTF-8 encoding.")
            return False
    return True


def get_args():
    # parse args
    parser = ArgumentParser(prog='zdict')

    parser.add_argument(
        'words',
        metavar='word',
        type=str,
        nargs='*',
        help='Words for searching its translation'
    )

    parser.add_argument(
        "-v", "--version",
        action="version",
        version='%(prog)s-' + constants.VERSION
    )

    parser.add_argument(
        "-d", "--disable-db-cache",
        default=False,
        action="store_true",
        help="Temporarily not using the result from db cache.\
              (still save the result into db)"
    )

    parser.add_argument(
        "-t", "--query-timeout",
        type=float,
        default=5.0,
        action="store",
        help="Set timeout for every query. default is 5 seconds."
    )

    def positive_int_only(value):
        ivalue = int(value)
        if ivalue <= 0:
            raise ArgumentTypeError(
                "%s is an invalid positive int value" % value
            )
        return ivalue

    parser.add_argument(
        "-j", "--jobs",
        type=positive_int_only,
        nargs="?",
        default=0,     # 0: not using, None: auto, N (1, 2, ...): N jobs
        action="store",
        help="""
            Allow N jobs at once.
            Do not pass any argument to use the number of CPUs in the system.
        """
    )

    parser.add_argument(
        "-sp", "--show-provider",
        default=False,
        action="store_true",
        help="Show the dictionary provider of the queried word"
    )

    parser.add_argument(
        "-su", "--show-url",
        default=False,
        action="store_true",
        help="Show the url of the queried word"
    )

    available_dictionaries = list(dictionary_map.keys())
    available_dictionaries.append('all')
    parser.add_argument(
        "-dt", "--dict",
        default="yahoo",
        action="store",
        choices=available_dictionaries,
        metavar=','.join(available_dictionaries),
        help="""
            Must be seperated by comma and no spaces after each comma.
            Choose the dictionary you want. (default: yahoo)
            Use 'all' for qureying all dictionaries.
            If 'all' or more than 1 dictionaries been chosen,
            --show-provider will be set to True in order to
            provide more understandable output.
        """
    )

    parser.add_argument(
        "-ld", "--list-dicts",
        default=False,
        action="store_true",
        help="Show currently supported dictionaries."
    )

    parser.add_argument(
        "-V", "--verbose",
        default=False,
        action="store_true",
        help="Show more information for the queried word.\
        (If the chosen dictionary have implemented verbose related functions)"
    )

    parser.add_argument(
        "-c", "--force-color",
        default=False,
        action="store_true",
        help="Force color printing (zdict automatically disable color printing \
        when output is not a tty, use this option to force color printing)"
    )

    parser.add_argument(
        '--dump', dest='pattern',
        nargs='?',
        default=None, const=r'^.*$',
        help='Dump the querying history, can be filtered with regex'
    )

    parser.add_argument(
        "-D", "--debug",
        default=False,
        action="store_true",
        help="Print raw html prettified by BeautifulSoup for debugging."
    )

    return parser.parse_args()


def set_args(args):
    if args.force_color:
        utils.Color.set_force_color()

    args.dict = args.dict.split(',')

    if 'all' in args.dict:
        args.dict = tuple(dictionary_map.keys())
    else:
        # Uniq and Filter the dict not in supported dictionary list then sort.
        args.dict = sorted(set(d for d in args.dict if d in dictionary_map))

    if len(args.dict) > 1:
        args.show_provider = True

    return args


def lookup_string_wrapper(dict_class, word, args):
    import sys
    if args.force_color:
        utils.Color.set_force_color()
    else:
        utils.Color.set_force_color(sys.stdout.isatty())

    dictionary = dict_class(args)
    f = StringIO()
    with redirect_stdout(f):
        dictionary.lookup(word)
    return f.getvalue()


def init_worker():
    # When -j been used, make subprocesses ignore KeyboardInterrupt
    # for not showing KeyboardInterrupt traceback error message.
    import signal
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def normal_mode(args):
    if args.jobs == 0:
        # user didn't use `-j`
        for word in args.words:
            for d in args.dict:
                zdict = dictionary_map[d](args)
                zdict.lookup(word)
    else:
        # user did use `-j`
        # If processes is None, os.cpu_count() is used.
        pool = Pool(args.jobs, init_worker)

        for word in args.words:
            futures = [
                pool.apply_async(lookup_string_wrapper,
                                 (dictionary_map[d], word, args))
                for d in args.dict
            ]
            results = [i.get() for i in futures]
            print(''.join(results))

    easter_eggs.lookup_pyjokes(word)


class MetaInteractivePrompt():
    def __init__(self, args):
        self.args = args
        self.dicts = tuple(
            dictionary_map[d](self.args) for d in self.args.dict
        )
        self.dict_classes = tuple(dictionary_map[d] for d in self.args.dict)

        if self.args.jobs == 0:
            # user didn't use `-j`
            self.pool = None
        else:
            # user did use `-j`
            # If processes is None, os.cpu_count() is used.
            self.pool = Pool(self.args.jobs, init_worker)

    def __del__(self):
        del self.dicts

    def prompt(self):
        user_input = input('[zDict]: ').strip()

        if user_input:
            if self.pool:
                futures = [
                    self.pool.apply_async(lookup_string_wrapper,
                                          (d, user_input, self.args))
                    for d in self.dict_classes
                ]
                results = [i.get() for i in futures]
                print(''.join(results))
            else:
                for dictionary_instance in self.dicts:
                    dictionary_instance.lookup(user_input)
        else:
            return

    def loop_prompt(self):
        while True:
            self.prompt()


def interactive_mode(args):
    # configure readline and completer
    readline.parse_and_bind("tab: complete")
    readline.set_completer(DictCompleter().complete)

    zdict = MetaInteractivePrompt(args)
    zdict.loop_prompt()


def execute_zdict(args):
    if args.list_dicts:
        for provider in sorted(
            dictionary_map,
            key=lambda x: {'yahoo': 0, 'pyjokes': 2}.get(x, 1)
        ):
            print(
                '{}: {}'.format(
                    provider,
                    dictionary_map[provider](args).title
                )
            )
        exit()

    if args.pattern:
        for word in dump(pattern=args.pattern):
            print(word)
        exit()

    try:
        if args.words:
            normal_mode(args)
        else:
            interactive_mode(args)
    except (KeyboardInterrupt, EOFError):
        print()
        return


def main():
    if user_set_encoding_and_is_utf8():
        check_zdict_dir_and_db()

        global dictionary_map
        dictionary_map = get_dictionary_map()

        args = get_args()
        args = set_args(args)

        execute_zdict(args)
    else:
        exit()



import json
import unicodedata  # to detect Unicode category

from zdict.dictionary import DictBase
from zdict.exceptions import QueryError, NotFoundError
from zdict.models import Record


class MoeDict(DictBase):

    API = 'https://www.moedict.tw/uni/{word}'

    @property
    def provider(self):
        return 'moe'

    @property
    def title(self):
        return '萌典'

    def _get_url(self, word) -> str:
        return self.API.format(word=word)

    def show(self, record: Record):
        content = json.loads(record.content)

        # print word
        self.color.print(content.get('title', ''), 'yellow')

        for word in content.get('heteronyms', ''):

            # print pronounce
            for key, display in (
                ('bopomofo', '注音'),
                ('bopomofo2', '注音二式'),
                ('pinyin', '漢語拼音')
            ):
                self.color.print(display, end='')
                self.color.print(
                    '[' + word.get(key, '') + ']',
                    'lwhite',
                    end=' ',
                )
            print()
            print()

            # print explain
            for count, explain in enumerate(word.get('definitions', '')):

                self.color.print(
                    '{order}. {text}'.format(
                        order=count + 1,
                        text=explain.get('def', '')
                    ),
                )

                if explain.get('synonyms'):
                    self.color.print(
                        '同義詞: {text}'.format(text=explain['synonyms']),
                        'magenta',
                        indent=2,
                    )

                if explain.get('antonyms'):
                    self.color.print(
                        '反義詞: {text}'.format(text=explain['antonyms']),
                        'magenta',
                        indent=2,
                    )

                for example in explain.get('example', ''):
                    self.color.print(
                        example,
                        'indigo',
                        indent=2,
                    )

                for quote in explain.get('quote', ''):
                    self.color.print(
                        '[引用] {text}'.format(text=quote),
                        'green',
                        indent=2,
                    )

                print()

    def query(self, word: str):
        try:
            content = self._get_raw(word)
        except QueryError as exception:
            raise NotFoundError(exception.word)

        record = Record(
            word=word,
            content=content,
            source=self.provider,
        )

        return record


def is_other_format(char):
    return unicodedata.category(char) != 'Cf'


def remove_cf(data):
    return ''.join(filter(is_other_format, data))


def clean(data, clean_cf=False):
    '''
    Clean the word segmentation

    remove "`~" and things in Unicode 'Cf' category
    '''
    data = data.translate(str.maketrans('', '', '`~'))
    if clean_cf:
        return remove_cf(data)
    else:
        return data


class MoeDictTaiwanese(DictBase):

    API = 'https://www.moedict.tw/t/{word}.json'

    @property
    def provider(self):
        return 'moe-taiwanese'

    @property
    def title(self):
        return '萌典（臺）'

    def _get_url(self, word) -> str:
        return self.API.format(word=word)

    def show(self, record: Record):
        content = json.loads(record.content)

        # print word
        self.color.print(clean(content.get('t', '')), 'yellow')

        for word in content.get('h', ''):

            # print pronounce
            for key, display in (
                # TODO: where is bopomofo ?
                ('T', '臺羅拼音'),      # Tailo
            ):
                self.color.print(display, end='')
                self.color.print(
                    '[' + word.get(key, '') + ']',
                    'lwhite',
                    end=' ',
                )

            print()
            print()

            # print explain
            for count, explain in enumerate(word.get('d', '')):

                self.color.print('{order}. '.format(order=count + 1), end='')
                type = clean(explain.get('type', ''))
                if type:
                    self.color.print(
                        '[' + type + ']',
                        'lgreen',
                        end=' ',
                    )

                self.color.print(clean(explain.get('f', '')), end='')

                for example in explain.get('e', ''):
                    self.color.print(
                        clean(example, True),
                        'indigo',
                        indent=2,
                    )

                print()

            print()

    def query(self, word: str):
        try:
            content = self._get_raw(word)
        except QueryError as exception:
            raise NotFoundError(exception.word)

        record = Record(
            word=word,
            content=content,
            source=self.provider,
        )

        return record



from pytest import raises
from unittest.mock import patch, Mock

from zdict.completer import DictCompleter


@patch('zdict.completer.db')
def test_DictCompleter_db_connect(db):
    '''
    Test case for DictCompleter init/del

    DictCompleter will connect to db when __init__,
    and close db when __del__.
    '''
    # init
    completer = DictCompleter()
    assert db.get_conn.called

    # del
    del completer
    assert db.close.called


@patch('zdict.completer.Record')
@patch('zdict.completer.db')
def test_DictCompleter_complete(db, Record):
    '''
    Test case for DictCompleter.complete implemented with iterator
    '''
    # Assume our db contains following words
    Record.select().where.return_value = [
        Mock(word='apple'),
        Mock(word='apply')]

    completer = DictCompleter()

    # init query
    ret = completer.complete('a', state=0)
    assert ret == 'apple'

    ret = completer.complete('a', state=1)
    assert ret == 'apply'

    with raises(StopIteration):
        completer.complete('a', state=2)



from pytest import raises
from unittest.mock import Mock, patch

from zdict.dictionaries.urban import UrbanDict
from zdict.exceptions import NotFoundError
from zdict.models import Record
from zdict.zdict import get_args


class TestUrbanDict:
    def setup_method(self, method):
        self.dict = UrbanDict(get_args())

    def teardown_method(self, method):
        del self.dict

    def test_provider(self):
        assert self.dict.provider == 'urban'

    def test__get_url(self):
        uri = 'http://api.urbandictionary.com/v0/define?term=mock'
        assert self.dict._get_url('mock') == uri

    def test_query_notfound(self):
        notfound_payload = '''
            {"tags":[],"result_type":"no_results","list":[],"sounds":[]}
        '''
        self.dict._get_raw = Mock(return_value=notfound_payload)

        with raises(NotFoundError):
            self.dict.query('mock')

        self.dict._get_raw.assert_called_with('mock')

    @patch('zdict.dictionaries.urban.Record')
    def test_query_normal(self, Record):
        self.dict._get_raw = Mock(return_value='{"mock": true}')
        self.dict.query('mock')
        Record.assert_called_with(
            word='mock',
            content='{"mock": true}',
            source='urban'
        )

    def test_show(self):
        content = '''
        {
            "list": [
                {
                    "word": "mock",
                    "definition": "Mock",
                    "example": "..."
                }
            ]
        }
        '''
        r = Record(word='mock', content=content, source='urban')

        # god bless this method, hope that it do not raise any exception
        self.dict.show(r)



from pytest import raises
from unittest.mock import Mock, patch

from zdict.dictionaries.yandex import YandexDict, API_KEY
from zdict.exceptions import NotFoundError, QueryError
from zdict.models import Record
from zdict.zdict import get_args


class TestYandexDict:
    def setup_method(self, method):
        self.dict = YandexDict(get_args())

    def teardown_method(self, method):
        del self.dict

    def test__get_url(self):
        url = 'https://translate.yandex.net/api/v1.5/tr.json/translate?' \
              'key={}&text=дом&lang=ru-en'
        assert url.format(API_KEY) == self.dict._get_url('дом')

    def test_provider(self):
        assert self.dict.provider == 'yandex'

    def test_query_timeout(self):
        self.dict._get_raw = Mock(side_effect=QueryError('дом', 404))

        with raises(NotFoundError):
            self.dict.query('дом')

        self.dict._get_raw.assert_called_with('дом')

    @patch('zdict.dictionaries.yandex.Record')
    def test_query_normal(self, Record):
        content = '{"code":200,"lang":"ru-en","text":["house"]}'
        self.dict._get_raw = Mock(return_value=content)
        self.dict.query('дом')
        Record.assert_called_with(word='дом', content=content, source='yandex')

    def test_show(self):
        content = '''
        {
            "code": 200,
            "lang": "ru-en",
            "text": ["house"]
        }
        '''
        r = Record(word='дом', content=content, source=self.dict.provider)

        # god bless this method, hope that it do not raise any exception
        self.dict.show(r)



import json

from zdict.dictionary import DictBase
from zdict.exceptions import NotFoundError, QueryError
from zdict.models import Record


# The daily request limit is 1,000,000 characters
# The monthly limit is 10,000,000 characters
API_KEY = 'trnsl.1.1.20170826T075621Z.' \
          '6dfcaff242c6caa8.e2b9cf136d451d9d6eb69516ec97b827e8c8229b'


class YandexDict(DictBase):
    """
    Docs:

    * https://tech.yandex.com/translate/
    * https://tech.yandex.com/dictionary/

    """

    # Change the url below to the API url of the new dictionary.
    # Need to keep the `{word}` for `_get_url()` usage.
    # TODO: support different translate direction
    # TODO: use Dictionary API
    API = 'https://translate.yandex.net/api/v1.5/tr.json/translate?' \
          'key={api_key}&text={word}&lang=ru-en'

    status_code = {
        200: 'Operation completed successfully',
        401: 'Invalid API key',
        402: 'Blocked API key',
        404: 'Exceeded the daily limit on the amount of translated text',
        413: 'Exceeded the maximum text size',
        422: 'The text cannot be translated',
        501: 'The specified translation direction is not supported',
    }

    @property
    def provider(self):
        return 'yandex'

    @property
    def title(self):
        return 'Yandex Translate'

    def _get_url(self, word) -> str:
        return self.API.format(api_key=API_KEY, word=word)

    def show(self, record: Record):
        content = json.loads(record.content)

        self.color.print(record.word, 'yellow')
        print()

        for index, data in enumerate(content['text']):
            self.color.print('{}. {}'.format(index + 1, data), 'org')

        print()

    def query(self, word: str):
        try:
            content = self._get_raw(word)
        except QueryError as exception:
            raise NotFoundError(exception.word)

        content_json = json.loads(content)

        status = content_json.get('code')
        if status != 200:
            # https://tech.yandex.com/translate/doc/dg/reference/translate-docpage/#codes
            message = self.status_code.get(
                status,
                'Some bad thing happened with Yandex'
            )
            print('Yandex: ' + message)
            raise NotFoundError(word)

        record = Record(
            word=word,
            content=content,
            source=self.provider,
        )
        return record
